(* Unit Kittycat v1.1 coded by Doug Good  09/27/00  Last updated 9/30/00

   Ok, so there's this big company "DC" who wants to give out free
   hardware, use it to collect and resell info on you, and then has
   the gall to say that you can't write your own programs to use
   the hardware...  well, SCREW 'EM!.  This unit will allow you to
   fully interact with those freebie handheld kitty cat barcode readers
   that Radio Shack is giving away.  Not only can you decode the output
   from those scanners (aka "Cat Poop") with this unit, but you can
   encode data into "Cat Poop" form too!

   Here's an example of raw Cat Poop:
   .C3nZC3nZC3nXDNnZDxrWC3nX.fHmc.C3DYDxT6ChnYCxf3.

   Here's the decoded version:
   000000002500673002  UPA  041689301224
    Serial Number     Type    Barcode

   How do you get from Cat Poop to useful data?  Here's the short version:
   Separate the Poop into 3 sections, delimited by the periods (Serial number,
   Barcode Type, and Barcode Value).  Divide each section into 4 character
   pieces.  Decode each piece using Base 64, with a modified translation table
   (uppercase and lowercase characters are swapped).  The decode leaves you
   with 3 characters of ASCII data.  Xor each character by the magic number
   of 67, and you have your decoded data. Repeat for each group of 4 characters
   in the Cat Poop string.  Thanks to Jean-Philippe Sugarbroad who, I'm told,
   hacked out this algorithm originally...

   --DG email: dgoodpa@hotmail.com

   REVISION HISTORY

   v1.1  - Added "MakeURL" function to facilitate sending a barcode to a URL server
         - Added "Delimiter" constant and removed a zillion instances of '.'  :)
         - Updated the Base64 Translation Table, changing the last character from
           '/' to a '-'.  There seems to be conflicting ideas as to what the last two
           characters in the translation table should be.  See this page for more
           info: http://www.fluent-access.com/wtpapers/cuecat/base64.html#final
         - ChangeSerial function was returning really wrong results... rather than
           returning valid Cat Poop, the result was half Poop and half decoded data.
           Whoops! Fixed.
   v1.0 Initial Release

*)

unit kittycat;

interface

// High level functions that get the job done
Function ExtractSerial(CatPoop:String):string;
Function ExtractCodeType(CatPoop:String):string;
Function ExtractBarcode(CatPoop:String):string;
Function MakeCatPoop(Serial,CodeType,Barcode:String):string;
Function ChangeSerial(Catpoop,NewSerial : string) : string;
Function MakeURL(AServer,AuthCode,CatPoop: string; FakeSerial: boolean) :string;

// Lower -level functions that are the meat of this unit
Function ReverseCase(str:string):string;
Function XorStr (astr:string;xval:integer):string;
Function DecodeOctet(octet:string):string;
Function EncodeTriplet(triplet:string):string;
Function DecodeString(codestring:string):string;
Function EncodeString(plaintext:string):string;
Function Base64Decode(Value : String) : String;
Function Base64Encode(Value : String) : String;

//Helper functions for the above routines...
Function sLeft(cString : String; iChars : Integer) : string;
Function sRight(cString : String; iChars : Integer) : string;
Function StripChar(var Data : string; OneChar : Char) : string;
Function ParseCount( cString: STRING; cChar: CHAR ): INTEGER;
Function Parse( cString: STRING; nIndex: INTEGER; cChar: CHAR ): STRING;

// NOTE: A standard BASE 64 table has the caps before the lowercase, not so with the Cat...
Const Base64Table = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789+-';
      MagicXorVal = 67; // The Cat's "Magic Number"
      Delimiter = '.';  // Cat Poop delimiter character

implementation

(*********************************************************************
   Function ExtractSerial
   Returns the decoded serial number from a given Cat Poop string
**********************************************************************)
Function ExtractSerial(CatPoop:String):string;
begin
  If CatPoop[1] = Delimiter then Delete(CatPoop, 1,1); // Kill first '.'
  Result := decodestring(Parse(CatPoop,1,Delimiter));
end; {ExtractSerial}

(********************************************************************
   Function ExtractCodeType
   Returns the decoded barcode type from a given Cat Poop string
*********************************************************************)
Function ExtractCodeType(CatPoop:String):string;
begin
  If CatPoop[1] = Delimiter then Delete(CatPoop, 1,1);
  Result := decodestring(Parse(CatPoop,2,Delimiter));
end; {ExtractCodeType}

(*********************************************************************
   Function ExtractBarcode
   Returns the decoded barcode from a given Cat Poop string
**********************************************************************)
Function ExtractBarcode(CatPoop:String):string;
begin
  If CatPoop[1] = Delimiter then Delete(CatPoop, 1,1);
  Result := decodestring(Parse(CatPoop,3,Delimiter));
end; {ExtractBarCode}


(*********************************************************************
   Function MakeCatPoop
   Returns an encoded string in the same format the Cat does, when
   given a serial number, code type and barcode number.
**********************************************************************)
Function MakeCatPoop(Serial,CodeType,Barcode:String):string;
begin
 result := Delimiter+EncodeString(serial)+Delimiter+EncodeString(CodeType)+
              Delimiter+EncodeString(barcode)+Delimiter;
end;

(*********************************************************************
   Function ChangeSerial
   Returns a new string of Cat Poop with the specified serial number.
   Each cat has a unique hardware serial number that is encoded in it's
   Poop.  This function will strip that hardware serial out of the Poop
   and replace it with any serial number you want.  For privacy reasons,
   you should use this before sending any Cat Poop to a URL server. "DC"
   might not like it, but hey, tough kitties!
**********************************************************************)
Function ChangeSerial(CatPoop,NewSerial : string) : string;
begin
  result := Delimiter+EncodeString(NewSerial)+Delimiter+parse(CatPoop,3,Delimiter)
            +Delimiter+parse(CatPoop,4,Delimiter)+Delimiter;
end;

(*********************************************************************
   Function MakeURL
   Returns a URL which can be used to retrieve info when a barcode is
   scanned in.  Server is the name of the server in format "a.xxxx.com",
   authorization code is typically a 22 character text string, but can
   be pretty much anything, apparently. Cat Poop is that string of
   excrement from your "feline shaped scanner".  (Yes, I'm taking great
   pains not to mention any specific companies,trademarks, or products
   in this code...) If Fake serial is true then we'll substitute our "special"
   serial number for the one that's hard coded in the cat.

   Returned URL format will look like this:
   http://[server]/CRQ/1..[activation code].04.[modified cat poop].0
*********************************************************************)
Function MakeURL(AServer,AuthCode,CatPoop: string; FakeSerial: boolean) :string;
begin
  if FakeSerial then //let's protect our privacy
   begin
     CatPoop := ChangeSerial(CatPoop,'000000000000031337');
   end;
  result := 'http://'+Aserver+'/CRQ/1..'+AuthCode+'.04'+reversecase(CatPoop)+'0';
end;


(*********************************************************************
   Function ReverseCase
   Returns a string with all alphabetic characters reversed in case.
   This function is necessary when sending Cat Poop to one of the
   servers.  For some reason, they encoded the Cat Poop with a case-
   reversed Base 64 algorithm, but the case must be returned to "normal"
   before being sent to a server.  This function does just that.
**********************************************************************)
function ReverseCase(str:string):string;
var x : integer;
    ch : char;
begin
  result := '';
  setlength(result,length(str));
  for x := 1 to length(str) do
   begin
      ch := str[x];
      if (ch in ['A'..'Z']) then
        ch := chr(ord(ch)+32)
      else
        if (ch in ['a'..'z']) then
         ch := chr(ord(ch)-32);
      result[x] := ch;
   end;
end;

(********************************************************************
   Function XorStr
   returns a string XOR'ed by a specified value
*********************************************************************)
Function XorStr (astr:string;xval:integer):string;
var x: integer;
    newstr: string;
begin
   setlength(newstr,length(astr));
   for x := 1 to length(astr) do
    begin
      newstr[x] := chr(ord(astr[x]) xor xval);
    end;
   result := newstr;
end;

(*********************************************************************
   DecodeOctet
   returns 1 octet decoded, the basis for building Cat Poop decoder
   functions.
**********************************************************************)
function DecodeOctet(octet:string):string;
begin
  result := xorstr(base64Decode(Octet),MagicXorVal);
end;

(**********************************************************************
   EncodeTriplet
   Given 3 ASCII characters, returns a coded Octet
   Used for building Cat Poop encoder functions
***********************************************************************)
function EncodeTriplet(triplet:string):string;
begin
  result := base64Encode(xorstr(triplet,MagicXorVal));
end;



(***********************************************************************
   function DecodeString:
   Returns decoded string, given the encoded string
   NOTE:  string must NOT contain "." delimiters!!
   Length of coded string should be evenly divisible by 4.
   Example Input:  C3nZC3nZC3nXDNnZDxrWC3nX
   Example Output: 000000002500673002
************************************************************************)
function DecodeString(codestring:string):string;
Const StrSize = 4;
var TotalOctets : integer;
    CurrentPos :Integer;
    x : integer;
    PString:string;
begin
    Pstring := '';
    CurrentPos := 1;
    TotalOctets := length(codestring) div StrSize; // how many octets in string?
    for x := 0 to TotalOctets-1 do
     begin
      Pstring := Pstring+decodeOctet(Copy(CodeString, CurrentPos,StrSize));
      currentpos := currentpos+StrSize;
     end;
    result := PString;
end;

(**********************************************************************
   function EncodeString:
   Given a plaintext string of length evenly divisible by 3, an encoded
   string will be returned.  Does NOT add delimiters.
   Example Input :  000000002500673002
   Example Output:  C3nZC3nZC3nXDNnZDxrWC3nX
***********************************************************************)
function EncodeString(plaintext:string):string;
Const StrSize = 3;
var TotalTriplets : integer;
    CurrentPos :Integer;
    x : integer;
    CString:string;
begin
    Cstring := '';
    CurrentPos := 1;
    TotalTriplets := length(plaintext) div StrSize; // how many triplets in string?
    for x := 0 to TotalTriplets-1 do
     begin
      cstring := cstring+EncodeTriplet(Copy(plaintext, CurrentPos, StrSize));
      currentpos := currentpos+StrSize;
     end;
    result := CString;
end;

// Decodes Base 64.  Not mine; I found this on the net...
// Came from the now defunct Beach Dog Software company.
Function Base64Decode(Value : String) : String;
Var AIn   : Array[1..4] Of Byte;
    AOut  : Array[1..3] Of Byte;
    AWork : Array[1..3] Of Byte;
    I : Integer;
    C : Integer;
Begin
   Result := '';
   I := 1;
   While I < Length(Value) Do
      Begin
         C := 3;
         FillChar(AWork, SizeOf(AWork), #0);
         FillChar(AOut, SizeOf(AWork), #0);
         AIn[1] := Byte(Pos(Value[I],Base64Table)-1);
         AIn[2] := Byte(Pos(Value[I+1],Base64Table)-1);
         AIn[3] := Byte(Pos(Value[I+2],Base64Table)-1);
         AIn[4] := Byte(Pos(Value[I+3],Base64Table)-1);
         If Value[I+3]='=' Then
            Begin
               C := 2;
               AIn[4] := 0;
               If Value[I+2]='=' Then
                  Begin
                     C := 1;
                     AIn[3] := 0;
                  End;
            End;
         AWork[2] := Byte(AIn[1] Shl 2);
         AWork[3] := Byte(AIn[2] Shr 4);
         AOut[1] := Byte(AWork[2] Or AWork[3]);
         AWork[2] := Byte(AIn[2] Shl 4);
         AWork[3] := Byte(AIn[3] Shr 2);
         AOut[2] := Byte(AWork[2] Or AWork[3]);
         AWork[2] := Byte(AIn[3] Shl 6);
         AOut[3] := Byte(AWork[2] Or AIn[4]);
         Result := Result + Char(AOut[1]);
         If C > 1 Then
            Result := Result + Char(AOut[2]);
         If C > 2 Then
            Result := Result + Char(AOut[3]);
         Inc(I, 4);
      End;
End;

// Encodes Base 64.  Not mine; I found this on the net...
// Came from the now defunct Beach Dog Software company.
Function Base64Encode(Value : String) : String;
Var AIn   : Array[1..3] Of Byte;
    AOut  : Array[1..4] Of Byte;
    AWork : Array[1..3] Of Byte;
    I : Integer;
    O : LongInt;
Begin
   Result := '';
   I := 1;
   O := Length(Value);
   Case Length(Value) Mod 3 Of
      1 : Value := Value + #0 + #0;
      2 : Value := Value + #0;
   End;
   While I < Length(Value) Do
      Begin
         AIn[1] := Byte(Value[I]);
         AIn[2] := Byte(Value[I+1]);
         AIn[3] := Byte(Value[I+2]);

         AOut[1] := Byte(AIn[1] Shr 2);
         AWork[1] := Byte(AIn[1] Shl 4);
         AWork[2] := Byte(AWork[1] And $30);
         AWork[3] := Byte(AIn[2] Shr 4);
         AOut[2] := Byte(AWork[2] Or AWork[3]);
         AWork[1] := Byte(AIn[2] Shl 2);
         AWork[2] := Byte(AWork[1] And $3C);
         AWork[3] := Byte(AIn[3] Shr 6);
         AOut[3] := Byte(AWork[2] Or AWork[3]);
         AOut[4] := Byte(AIn[3] And $3F);

         Inc(I, 3);
         Result := Result + Base64Table[AOut[1]+1] + Base64Table[AOut[2]+1] + Base64Table[AOut[3]+1] + Base64Table[AOut[4]+1];
      End;
   If O Mod 3 > 0 Then
      Result[Length(Result)] := '=';
   If O Mod 3 = 1 Then
      Result[Length(Result)-1] := '=';
End;

(*******************************************************************)
(*******************************************************************
  The remainder of the functions in this unit are used to parse
  strings, and simply make life easier.  I didn't write these, but
  included them so everything would be in one nice, neat package.
 *******************************************************************)
(*******************************************************************)

Function sLeft(cString : String; iChars : Integer) : string;
begin
     sLeft := copy(cString,1,iChars);
end;

Function sRight(cString : String; iChars : Integer) : string;
begin
     sright := copy(cString,Length(cString)-ichars+1,iChars);
end;

function StripChar(var Data : string; OneChar : Char) : string;
begin
   repeat
     delete(data,pos(Onechar,data),1);
   until pos(Onechar,data) = 0;
   StripChar := Data;
end;

{*****************************************************************************
 * Function ...... ParseCount()
 * Purpose ....... To count the number of tokens in a string
 * Parameters .... cString      String to count tokens in
 *                 cChar        Token separator
 * Returns ....... Number of tokens in <cString>
 * Notes ......... Uses function StripChar
 * Author ........ Martin Richardson
 *****************************************************************************}
FUNCTION ParseCount( cString: STRING; cChar: CHAR ): INTEGER;
var oldlen,newlen : integer;
BEGIN
     oldlen := length(cString);
     newlen := length(stripchar(cString,cChar));
     ParseCount := oldlen - newlen;
END;

{*****************************************************************************
 * Function ...... Parse()
 * Purpose ....... To parse out tokens from a string
 * Parameters .... cString      String to parse
 *                 nIndex       Token number to return
 *                 cChar        Token separator
 * Returns ....... Token <nIndex> extracted from <cString>
 * Notes ......... If <nIndex> is greater than the number of tokens in
 *                 <cString> then a null string is returned.
 *               . Uses function sLeft, sRight, and ParseCount
 * Author ........ Martin Richardson
 * Date .......... September 30, 1992
 *****************************************************************************}
FUNCTION Parse( cString: STRING; nIndex: INTEGER; cChar: CHAR ): STRING;
VAR
   i: INTEGER;
   cResult: STRING;
BEGIN
     cString := cString + cChar;
     IF nIndex > (ParseCount( cString, cChar )) THEN
        cResult := ''
     ELSE BEGIN
          FOR i := 1 TO nIndex DO BEGIN
              cResult := sLeft( cString, POS( cChar, cString ) - 1 );
              cString := sRight(cString, LENGTH(cString) - POS(cChar, cString));
          END { Next I };
     END { IF };
     Parse := cResult;
END;


end.
